import numpy as np
from matplotlib import pyplot as plt
from matplotlib import animation, rc

rc("text", usetex=True)


def animate(sol, title="", savepath=None, fig_size=(6, 6), variable=0):
    """
    Method that animates the time evolution of the defined system. Optionally saves an mp4 video file to
    'savepath'.

    Parameters
    ----------
    sol : np.array
        Array of shape either (#Variables, #Timesteps, #Dim1) or (#Variables, #Timesteps, #Dim1, #Dim2)
    title : string
        (Optional, default="") Give a title for the resulting animation
    savepath : string
        (Optional, default=None) Give a savepath for the animation
    fig_size : tuple
        (Optional, default=(6, 6)) Give a figure size for matplotlib
    variable : int
        (Optional, default=0) Gives the index of the variable to show (if there are more than one)

    Returns
    -------
    None

    """
    fig, ax = plt.subplots(figsize=fig_size)
    dimension = len(sol.shape) - 2

    if dimension == 2:
        sol_img = ax.imshow(sol[variable, 0])
    elif dimension == 1:
        sol_img, = ax.plot(sol[variable, 0])
    else:
        raise NotImplementedError("Only 1D and 2D solutions can be animated.")

    # This feels stupid (trying to avoid "if" in animation call)+
    def anime2d(i):
        plt.title(title + f" Timesteps: {i}")

        sol_img.set_data(sol[variable, i])
        return sol_img

    def anime1d(i):
        plt.title(title + f" Timesteps: {i}")

        sol_img.set_ydata(sol[variable, i])
        return sol_img

    # This feels even more stupid
    if dimension == 2:
        anime = anime2d
    elif dimension == 1:
        anime = anime1d

    anima = animation.FuncAnimation(fig, anime, frames=np.arange(0, len(sol[variable, :]), 1), interval=10,
                                    repeat=False)
    if savepath is not None:
        Writer = animation.writers["ffmpeg"]
        writer = Writer(fps=15, bitrate=1800)
        anima.save(savepath, writer=writer)
    else:
        plt.show()



import numpy as np
from scipy.integrate import solve_ivp
from .differentiation import DifferentialOperator, get_differential_operator


class AbstractTimeSimulator:
    """
    This abstract class provides basic time simulation functionality for ODEs and PDEs (1D or 2D) using Scipys initial
    value solver. It is intended to be extended with needed attributes and a right-hand-side function (rhs), It also
    provides methods for handling spatial differentiation and boundary conditions.
    """

    def __init__(self, ic, domain=None, bc=None, diff=None):
        """
        Basic constructor method. Provides spatial coordinates (if applicable) and records basic information of the
        problem. Also creates some protected attributes for differentiation methods.

        Parameters
        ----------
        ic : array-like
            array of initial conditions, should be in the shape (# Variables, ) for ODEs, (# Variables, # First
            spatial dimension) for 1D PDEs and (# Variables, # First spatial dimension, # Second spatial dimension) for
            2D PDEs
        domain : tuple
            physical domain of the problem, for ODE value is None (default), for PDE shape should be ((x_min, x_max), )
            in the 1D case and ((x_min, x_max), (y_min, y_max)) for 2D

        Attributes
        ----------
        To be extended
        """
        if diff == "pseudo_spectral" and bc != "periodic":
            bc = "periodic"
            print("Warning: Boundary condition forced to 'periodic' due to pseudo-spectral method!")

        assert diff is None or diff == "pseudo_spectral" or diff == "finite_difference", \
            "Unsupported differentiation method."

        self.n_variables = ic.shape[0]  # Number of variables of the problem
        self.domain = domain  # Passed domain as specified above
        self.initial_state = ic  # Passed initial conditions as stated above
        self._dimensions = None  # number of spatial dimension of the problem
        self._discret = None  # spatial discretization of the problem (if applicable)
        self.diff = None  # instance of DiffirentialOperator handling spatial differentiation
        self.solver = None  # to save the scipy solver object (maybe remove)
        self.sol = None  # to save time simulation
        self.sol_dot = None  # to save derivative of time simulation
        self.time = None  # time stamps of the simulation

        # Check provided domain to see for different use-cases: ODE, 1D PDE, 2D PDE
        if self.domain is None:
            self._mode = "ODE"
        else:
            self._mode = "PDE"
            self._dimensions = len(domain)
            self._discret = self.initial_state.shape[1]

            # Check to see if discretization along the dimensions match and also if domain matches initial conditions
            if self._dimensions == 2:
                try:
                    self.initial_state.shape[2]
                except IndexError:
                    raise ValueError("Dimension of domain and initial conditions do not match.")
                assert self._discret == self.initial_state.shape[2], "Discretization needs to be the same in either " \
                                                              "dimension."

        # All of the following enables use of spatial coordinates and derivatives in rhs function call
        if self._dimensions is None:
            pass
        elif self._dimensions == 1:
            self.Lx = self.domain[0][1] - self.domain[0][0]  # Absolute length of the spatial interval
            self.x = np.linspace(self.domain[0][0], self.domain[0][1], num=self._discret)  # spatial coordinates
            # create object handling spatial derivatives
            self.diff = DifferentialOperator({"Lx": self.Lx, "x": self.x, "n": self._discret},
                                             boundary=bc, method=diff)
        elif self._dimensions == 2:
            self.Lx = self.domain[0][1] - self.domain[0][0]  # Absolute length of the first spatial interval
            self.Ly = self.domain[1][1] - self.domain[1][0]  # Absolute length of the second spatial interval
            self.x = np.linspace(self.domain[0][0], self.domain[0][1], self._discret)  # spatial coordinates
            self.y = np.linspace(self.domain[1][0], self.domain[1][1], self._discret)
            self.x_mesh, self.y_mesh = np.meshgrid(self.x, self.y)
            # create object handling spatial derivatives
            self.diff = DifferentialOperator({"Lx": self.Lx, "x": self.x,
                                              "Ly": self.Ly, "y": self.y, "n": self._discret},
                                             boundary=bc, method=diff)
        else:
            raise NotImplementedError

    def rhs(self, t, u):
        """
        Abstract method for right-hand-side function of initial value problem. Per default available are:
        self.x (and self.y): spatial coordinates
        self.diff.d_dx(u, o): spatial derivative in first spatial component of order o
        self.diff.d_dy(u, o): spatial derivative in second spatial component of order o (if applicable)
        Other quantities should be computed in the constructor method when they are independent of u and t.

        Parameters
        ----------
        self
        t : float
            current time during simulation
        u : array-like
            current (unflattened) variables during time simulation

        Returns
        -------
        u_next : array-like
            values of variables for the next time step
        """
        pass

    def _flat_rhs(self, t, uin):
        """
        Helper method that corrects input and output dimension between rhs function and scipys ivp solver
        Parameters
        ----------
        self
        t : float
            time argument to pass along to the function
        uin : array-like
            variable values to pass along to the function

        Returns
        -------
        flat_u_next : array-like
            a flattend version of the values for the variables in the next time step
            """
        if self._dimensions is None:
            u = uin
        elif self._dimensions == 1:
            u = np.reshape(uin, (self.n_variables, self._discret))
        elif self._dimensions == 2:
            u = np.reshape(uin, (self.n_variables, self._discret, self._discret))
        else:
            raise NotImplementedError

        result = self.rhs(t, u)

        return result.flatten()

    def simulate(self, time_interval, method="DOP853", t_eval=None):
        # solve initial value problem
        self.solver = solve_ivp(self._flat_rhs, time_interval, self.initial_state.flatten(),
                                method=method, t_eval=t_eval)
        # save solution
        self.sol = [self.solver.y[:, i] for i in range(len(self.solver.y[0, :]))]
        # de-flatten output
        if self._dimensions is None:
            self.sol = np.array([np.reshape(item, (self.n_variables,)) for item in self.sol])
        elif self._dimensions == 1:
            self.sol = [np.reshape(item, (self.n_variables, self._discret)) for item in self.sol]
            self.sol = np.array(self.sol)
        elif self._dimensions == 2:
            self.sol = [np.reshape(item, (self.n_variables, self._discret, self._discret)) for item in self.sol]
            self.sol = np.array(self.sol)

        self.sol = np.array([self.sol[:, i] for i in range(self.sol.shape[1])])
        self.time = self.solver.t

        return self.sol

    def get_sol_dot(self, t=None, s=None):
        if t is None:
            time = self.time
        else:
            time = t
        if s is None:
            sol = self.sol
        else:
            sol = s

        assert time is not None, "Need to pass array of time stamps or perform a time simulation"
        assert sol is not None, "A solution array needs to be provided or a time simulation needs to be performed"

        d_dt = get_differential_operator(time, 1)
        self.sol_dot = d_dt(sol)
        return self.sol_dot





# from .version import __version__
# from .shout import shout_and_repeat
# from .add import my_add

# if somebody does "from somepackage import *", this is what they will
# be able to access:
# __all__ = [
#     'shout_and_repeat',
#     'my_add',
# ]


import numpy as np
from findiff import FinDiff


def get_differential_operator(array, order):
    return FinDiff(order, array)


class DifferentialOperator:
    """
    Class that handles numerical differentiation under certain boundary conditions.
    """

    def __init__(self, space_dict, boundary="periodic", method="pseudo_spectral"):

        n = space_dict["n"]  # get number of spatial points in each dimension
        self.p = 5  # set size of padding for finite difference method

        if "y" in space_dict.keys():
            self.dimensions = 2
            self.x = space_dict["x"]
            self.y = space_dict["y"]

            if method == "finite_difference":
                # for boundary conditions in finite differences
                x_pad = np.linspace(self.x[0] - self.p * (self.x[1] - self.x[0]),
                                    self.x[-1] + self.p * (self.x[1] - self.x[0]),
                                    num=n + 2 * self.p)

                y_pad = np.linspace(self.y[0] - self.p * (self.y[1] - self.y[0]),
                                    self.y[-1] + self.p * (self.y[1] - self.y[0]),
                                    num=n + 2 * self.p)

                # this is to accommodate for one and two dimension in the differentiation method and feels discouraged
                def unpad(u):
                    return u[self.p:-self.p, self.p:-self.p]

                self.d_dx = lambda u, o: self.finite_difference(u, 0, x_pad, order=o, boundary=boundary, unpad=unpad)
                self.d_dy = lambda u, o: self.finite_difference(u, 1, y_pad, order=o, boundary=boundary, unpad=unpad)
                self.dd_dxdy = lambda u, o: self.finite_difference(u, [0, 1], [x_pad, y_pad], order=o,
                                                                   boundary=boundary, unpad=unpad)

            elif method == "pseudo_spectral":

                # for pseudo-spectral method
                kx, ky = np.meshgrid(np.fft.fftfreq(n, space_dict["Lx"] / (n * 2 * np.pi)),
                                     np.fft.fftfreq(n, space_dict["Ly"] / (n * 2 * np.pi)))

                self.d_dx = lambda u, o: self.pseudo_spectral(u, kx, order=o, fft=np.fft.fft2, ifft=np.fft.ifft2)
                self.d_dy = lambda u, o: self.pseudo_spectral(u, ky, order=o, fft=np.fft.fft2, ifft=np.fft.ifft2)
                self.dd_dxdy = lambda u, o: self.d_dx(self.d_dy(u, o[1]), o[0])

            self.div = lambda u, o: self.d_dx(u, o) + self.d_dy(u, o)

        else:
            self.dimensions = 1
            self.x = space_dict["x"]

            if method == "finite_difference":
                # for boundary conditions in finite differences
                x_pad = np.linspace(self.x[0] - self.p * (self.x[1] - self.x[0]),
                                    self.x[-1] + self.p * (self.x[1] - self.x[0]),
                                    num=n + 2 * self.p)

                def unpad(u):
                    return u[self.p:-self.p]

                self.d_dx = lambda u, o: self.finite_difference(u, 0, x_pad, order=o, boundary=boundary, unpad=unpad)
                self.d_dy = None
                self.dd_dxdy = None

            elif method == "pseudo_spectral":
                kx = np.fft.fftfreq(n, space_dict["Lx"] / (n * 2 * np.pi))
                self.d_dx = lambda u, o: self.pseudo_spectral(u, kx, order=o, fft=np.fft.fft, ifft=np.fft.ifft)
                self.d_dy = None
                self.dd_dxdy = None
            self.div = self.d_dx

    def finite_difference(self, u, var, pad, order=1, boundary="periodic", unpad=None):
        if boundary == "periodic":
            bound = "wrap"
        elif boundary == "neumann":
            bound = "edge"
        else:
            raise NotImplementedError("Boundary condition not implemented.")

        u_padded = np.pad(u, [(self.p, self.p) for i in range(len(u.shape))], bound)
        if not(isinstance(var, int)):
            try:
                dn_dxn = FinDiff((var[0], pad[0], order[0]), (var[1], pad[1], order[1]))
            except TypeError:
                raise TypeError("Multivariate derivative requires lists or arrays for all (numeric) inputs")
        else:
            dn_dxn = FinDiff(var, pad, order)

        gradient = dn_dxn(u_padded)
        return unpad(gradient)

    def pseudo_spectral(self, u, k, order=1, fft=np.fft.fft2, ifft=np.fft.ifft2):
        return ifft((1j * k) ** order * fft(u))



import numpy as np
import scipy.integrate as integrate
import optuna
from gradient_free_optimizers import HillClimbingOptimizer
import nevergrad as ng
from hyperopt import fmin, hp, tpe
from tqdm import tqdm


# Maybe at some point add option for sklearns rigde regressor:
# from sklearn.linear_model import Ridge


def sequencing_threshold_ridge(lib, data, lamb, tol=1e-4, max_it=10, norm=2):
    """
    This reproduces the sequencing threshold ridge regression as it is implemented in PDE-FIND.

    Parameters
    ----------
    lib : np.array
        Time series of library terms as produced by a simulator instance via method create_library_time_series or of
        shape (# Variables, # Time Steps, # Library Terms, Dim1, Dim2(if applicable))
    data : np.array
        Time series to be fitted to (in shape (# Variable, # Time steps, Dim1, Dim2(if applicabe)))
    lamb : float
        Sparsity knob to promote sparsity in regularisation
    tol : float
        (Optional, default: 1e-4) Threshold below which values are cut in sequence of regression
    max_it : int
        (Optional, default: 10) Maximum number of iterations in sequence of regressions
    norm : int
        (Optional, default: 2) Give the norm with which to normalize data

    Returns
    -------
    np.array of weights for linear combination of library terms to best achieve data series

    """
    # Initialize weights and read the number of variables and the number of library terms
    w = []
    n_var = lib.shape[0]
    n_terms = lib.shape[1]

    # We try to find a set of weights for every variables right hand side
    for var in range(n_var):
        # Read and reshape the library terms for the current variable, here we flatten all except the first dimension
        x = np.reshape(lib[var], (lib[var].shape[0], np.prod(lib[var].shape[1:])))
        # Flatten the corresponding data time series
        y = data[var].flatten()
        # When normalization is set, save the scaling factor and normalize the library according to the parameter norm
        l_norm = None
        tol_loc = tol
        if norm != 0:
            l_norm = 1.0 / (np.linalg.norm(x, norm))
            x = l_norm * x
            tol_loc = tol_loc / l_norm
        if lamb != 0:
            # If a sparsity knob is given get the standard ridge estimate
            w_loc = np.linalg.lstsq(x.dot(x.T) + lamb * np.eye(n_terms), x.dot(y), rcond=None)[0]
        else:
            # else perform ordinary least squares fit
            w_loc = np.linalg.lstsq(x.T, y, rcond=None)[0]

        # Set the number of relevant terms to the number of all terms
        n_relevant = n_terms
        for i in range(max_it):
            # We find all the indices which hold values below our threshold and the remaining indices
            off_indices = np.where(np.abs(w_loc) < tol_loc)[0]
            new_on_indices = np.delete(np.arange(n_terms), off_indices)

            if n_relevant == len(new_on_indices):
                # Stop if the number of relevant terms does not change
                break
            else:
                # Update the number of relevant terms
                n_relevant = len(new_on_indices)

            on_indices = new_on_indices

            if len(on_indices) == 0:
                # If all indices have been tossed stop (if this happend during the first iteration return the value)
                if i == 0:
                    # w.append(w_loc)
                    break
                else:
                    break

            # Cut off all values to the corresponding indices
            w_loc[off_indices] = 0
            # Get knew ridge or least square estimate
            if lamb != 0:
                w_loc[on_indices] = np.linalg.lstsq(x[on_indices].dot(x[on_indices].T) +
                                                    lamb * np.eye(len(on_indices)), x[on_indices].dot(y), rcond=None)[0]
            else:
                w_loc[on_indices] = np.linalg.lstsq(x[on_indices].T, y, rcond=None)[0]

        # Normalize the weights
        if norm != 0:
            w_loc[np.where(np.abs(w_loc) < tol_loc)[0]] = 0
            w.append(np.multiply(l_norm, w_loc))
        else:
            w_loc[np.where(np.abs(w_loc) < tol_loc)[0]] = 0
            w.append(w_loc)

    return np.array(w)


def train_ridge(lib, data, lamb, thres, l0=None, split=0.7, seed=1234, max_it_train=10, max_it=10, norm=2):
    np.random.seed(seed)
    n_timesteps = lib.shape[2]
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    d_tol = float(thres)
    tol = thres
    tol_best = tol

    if l0 is None:
        l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
                  for i in range(x_test.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    w_best = np.array([np.linalg.lstsq(x_test[i].T, y_test[i], rcond=None)[0]
                       for i in range(x_test.shape[0])])
    error_best = error(w_best)

    pbar = tqdm(range(max_it_train))
    pbar.set_description(f"Training sequencing threshold ridge regression with "
                         f"{split * 100:.0f} % split")
    for i in pbar:
        w = sequencing_threshold_ridge(x_train, y_train, lamb, max_it=max_it, tol=tol, norm=norm)
        err = error(w)

        if err <= error_best:
            error_best = err
            tol_best = tol
            w_best = w
            tol = tol + d_tol
        else:
            tol = max([0, tol - 2 * d_tol])
            d_tol = 2 * d_tol / (max_it_train - i)
            tol = tol + d_tol
        pbar.set_postfix({"Threshold": tol_best, "Error": error_best})

    return np.array(w_best), error_best


def train_ridge_wip(lib, data, data_dot, time, lamb, thres, l0=None, seed=1234, max_it_train=10, max_it=10, norm=2,
                    error="BIC"):
    np.random.seed(seed)
    n_timesteps = lib.shape[2]
    # reshaping is done inconsistently
    # x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    # y = np.reshape(data_dot, (data_dot.shape[0], np.prod(data_dot.shape[1:])))
    x = lib
    y = data_dot

    d_tol = float(thres)
    tol = thres
    tol_best = tol

    if l0 is None:
        l0 = 0.001 * np.linalg.norm(np.linalg.cond(x.reshape((x.shape[0], x.shape[1], np.prod(x.shape[2:])))), 2)

    def error_sp(w_test):
        errors = [np.linalg.norm(data[i, 1:] - integrate.cumtrapz(x[i].T.dot(w_test[i]), time).T)
                  for i in range(x.shape[0])]

        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def error_bic(w_test):
        errors = np.array([np.linalg.norm(data[i, 1:] - integrate.cumtrapz(x[i].T.dot(w_test[i]), time).T)
                           for i in range(x.shape[0])])
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(w_test)
        bic = -2 * logL + p * np.log(len(time))
        return bic

    if error == "BIC":
        error = error_bic
    else:
        error = error_sp

    w_best = np.array([np.linalg.lstsq(x[i].reshape(x.shape[1], -1).T, y[i].flatten(), rcond=None)[0]
                       for i in range(x.shape[0])])
    error_best = error(w_best)

    pbar = tqdm(range(max_it_train))
    pbar.set_description(f"Training sequencing threshold ridge regression with integration error")
    for i in pbar:
        w = sequencing_threshold_ridge(x, y, lamb, max_it=max_it, tol=tol, norm=norm)
        err = error(w)

        if err <= error_best:
            error_best = err
            tol_best = tol
            w_best = w
            tol = tol + d_tol
        else:
            tol = max([0, tol - 2 * d_tol])
            d_tol = 2 * d_tol / (max_it_train - i)
            tol = tol + d_tol
        pbar.set_postfix({"Threshold": tol_best, "Error": error_best})

    return np.array(w_best), error_best


def optimize(lib, data, data_dot, time):
    def training_loss_bic(weights):
        errors = np.array([np.linalg.norm(data[i, 1:] - integrate.cumtrapz(lib[i].T.dot(weights[i]), time).T)
                           for i in range(lib.shape[0])])
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(weights)
        bic = -2 * logL + p * np.log(len(time))
        return bic

    def BIC(lamb):
        weights = sequencing_threshold_ridge(lib, data_dot, lamb)
        return training_loss_bic(weights)

    opt = fmin(BIC, space=hp.loguniform("l", -20, 0), algo=tpe.suggest, max_evals=20)
    print(f"Optimal lambda found: {opt['l']}")
    return sequencing_threshold_ridge(opt["l"])


def optimize_ridge_optuna(lib, data, split=0.7, seed=1234, trials=20, tol_range=[1e-4, 2], lamb_range=[0, 2]):
    np.random.seed(seed)
    # l0 = 0.001 * np.linalg.norm(np.linalg.cond(lib), 2)
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
                  for i in range(x_test.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def objective(trial):
        tol = trial.suggest_float("tol", tol_range[0], tol_range[1])
        # tol = 1.0
        lamb = trial.suggest_float("lamb", lamb_range[0], lamb_range[1])
        # lamb = 0.01
        w = sequencing_threshold_ridge(x_train, y_train, lamb, tol=tol)
        # errors = [np.linalg.norm(data[i].flatten() - lib[i].T.dot(w[i]).flatten(), 2)
        #          for i in range(lib.shape[0])]
        # error_test = np.sum(errors) + l0 * np.count_nonzero(w)# + 1 / np.sum([np.linalg.norm(wi, 2) for wi in w])
        error_test = error(w)
        return error_test

    study = optuna.create_study()
    study.optimize(objective, n_trials=trials)
    params = study.best_params
    lam_best = params["lamb"]
    # lam_best = 0.01
    tol_best = params["tol"]
    # tol_best = 1.0

    return sequencing_threshold_ridge(lib, data, lam_best, tol=tol_best)


def optimize_ridge_gfo(lib, data, split=0.7, seed=1234, trials=20, tol_range=[1e-4, 2], lamb_range=[0, 2]):
    np.random.seed(seed)
    # l0 = 0.001 * np.linalg.norm(np.linalg.cond(lib), 2)
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
                  for i in range(x_test.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def objective(para):
        tol = para["tol"]
        # tol = 1.0
        lamb = para["lamb"]
        # lamb = 0.01
        w = sequencing_threshold_ridge(x_train, y_train, lamb, tol=tol)
        # errors = [np.linalg.norm(data[i].flatten() - lib[i].T.dot(w[i]).flatten(), 2)
        #          for i in range(lib.shape[0])]
        # error_test = np.sum(errors) + l0 * np.count_nonzero(w)# + 1 / np.sum([np.linalg.norm(wi, 2) for wi in w])
        error_test = error(w)
        return - error_test

    search_space = {"tol": tol_range, "lamb": lamb_range}
    opt = HillClimbingOptimizer(search_space)
    opt.search(objective, n_iter=trials)

    lam_best, tol_best = opt.best_value

    return sequencing_threshold_ridge(lib, data, lam_best, tol=tol_best)


def optimize_ridge_optuna_wip(model, split=0.7, seed=1234, trials=20, tol_range=[1e-4, 2], lamb_range=[0, 2]):
    lib = model.simulator.library_over_time
    data = model.simulator.sol_dot
    time = model.simulator.time
    Nt = len(time)

    np.random.seed(seed)
    # l0 = 0.001 * np.linalg.norm(np.linalg.cond(lib), 2)
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        # errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
        #           for i in range(x_test.shape[0])]

        model.simulator.sigma = w_test
        y_estimate = model.simulator.simulate([time[0], time[-1]], t_eval=time)
        diff = ((model.sol - y_estimate).flatten()) ** 2
        logL = -Nt * np.log(diff.sum() / Nt) / 2

        s = w_test.flatten()
        s[np.abs(s != 0)] /= np.abs(s[s != 0])
        p = np.abs(s).sum()

        bic = -2 * logL + p * np.log(Nt)

        return bic

    def objective(trial):
        tol = trial.suggest_float("tol", tol_range[0], tol_range[1])
        # tol = 1.0
        lamb = trial.suggest_float("lamb", lamb_range[0], lamb_range[1])
        # lamb = 0.01
        w = sequencing_threshold_ridge(x_train, y_train, lamb, tol=tol)
        # errors = [np.linalg.norm(data[i].flatten() - lib[i].T.dot(w[i]).flatten(), 2)
        #          for i in range(lib.shape[0])]
        # error_test = np.sum(errors) + l0 * np.count_nonzero(w)# + 1 / np.sum([np.linalg.norm(wi, 2) for wi in w])
        error_test = error(w)
        return error_test

    study = optuna.create_study()
    study.optimize(objective, n_trials=trials)
    params = study.best_params
    lam_best = params["lamb"]
    # lam_best = 0.01
    tol_best = params["tol"]
    # tol_best = 1.0

    return sequencing_threshold_ridge(lib, data, lam_best, tol=tol_best)


def optimize_wip(model, w_init, trials=20, n_time=5):
    time = model.simulator.time[:n_time]
    Nt = len(time)

    def error(w_test):
        # errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
        #           for i in range(x_test.shape[0])]

        model.simulator.sigma = w_test
        y_estimate = model.simulator.simulate([time[0], time[-1]], t_eval=time)
        try:
            diff = ((model.sol - y_estimate).flatten()) ** 2
        except ValueError:
            diff = np.array([1e10])

        logL = -Nt * np.log(diff.sum() / Nt) / 2

        s = w_test.flatten()
        s[np.abs(s != 0)] /= np.abs(s[s != 0])
        p = np.abs(s).sum()

        bic = -2 * logL + p * np.log(Nt)

        return bic

    instru = ng.p.Instrumentation(ng.p.Array(init=w_init))
    optimizer = ng.optimizers.NGOpt(parametrization=instru, budget=trials)
    recommendation = optimizer.minimize(error)

    return recommendation.value[0][0]



import numpy as np
from numpy import math
from tsme.time_simulation import AbstractTimeSimulator
from .optimization import train_ridge_wip
from itertools import combinations_with_replacement as combinations_wr
from tabulate import tabulate


def _make_combination_string(order, n_variables):
    """
    This helper method gives a string containing all combinations of multiplying n_variables with one another and
    also the single elements.
    Example output for order=2 and n_variables=3:
    `np.array([np.ones(u[0].shape), u[0], u[1], u[2], u[0]*u[0], u[0]*x[1], u[0]*u[2], u[1]*u[1], u[1]*u[2], u[2]*u[2]])`

    Parameters
    ----------
    order : int
        Order up to which permutations should be included
    n_variables :
        Number of variables to combine

    Returns
    -------
    x_string : string
        String containing a 'one' element, the elements of an array u and products of all combinations of elements of u.

    """
    xstr = [f"u[{i}]" for i in range(n_variables)]
    x_string = "np.array([np.ones(u[0].shape)"
    for o in range(order + 1):
        comb = combinations_wr(xstr, o)
        for c in list(comb):
            if c == ():
                continue
            x_string += ", "
            for i, t in enumerate(c):
                if i == 0:
                    x_string += t
                else:
                    x_string += "*" + t
    x_string += "])"
    return x_string


class Model:
    def __init__(self, sol, time, phys_domain=None, bc="periodic", diff="finite_difference"):
        """
        This class is used to wrap the time simulation capabilities of the time_simulation submodule.
        It supplies a method to create a library of possible right-hand-side terms and creates an instance of a Time-
        simulator object used to perfom time simulations provided a vector `sigma` giving factors for linearly combining
        library terms for each variable. The overall geometry of the problem is extracted from the initial conditions
        provided during initialization.

        Parameters
        ----------
        sol : numpy.array
            Time series data
        time : numpy.array
            Array of points in time for each data point in time series data
        phys_domain : tuple
            (Optional, default=None) Physical domain size in shape e. g. ((Lx_min, Lx_max), (Ly_min, Ly_max))
            If None will set the domain size equal to the number of spatial discretization steps (if applicable)
        bc : string
            (optional, default="periodic") Sets boundary conditions (mainly for finite difference method). Options are
            "periodic" and "neumann"
        diff : string
            (optional, default="finite_difference") Sets differentiation method for spatial derivative. Options are
            "finite_difference" and "pseudo_spectral"
        """

        ic = sol[:, 0]
        self.initial_state = ic

        self.simulator = None
        self.sol = sol
        self.time = time

        self.n_variables = ic.shape[0]
        self.N = ic.shape[-1]
        self.dimension = len(ic[0].shape)
        # In case of ODE system
        if len(ic.shape) == 1:
            self.dimension = 0
            self.N = 0
        if phys_domain is None and self.dimension is not 0:
            self.domain = [(0, self.N) for i in range(self.dimension)]
        else:
            # assert self.dimension is not 0, "Cannot specify physical domain for ODE system."
            self.domain = phys_domain
        self.boundary_condition = bc
        self.differentiation_method = diff

        self.ode_order = None  # Maximum number of factors in products of permutations
        self.indices_ode_to_pde = None  # Maximum order of spatial derivative
        self.pde_order = None  # Maximum order of factors in products for spatial derivatives
        self.ode_count = 0
        self.pde_count = 0
        self.custom_count = 0
        self.ode_string = None
        self.pde_string = None
        self.custom_string = None
        self._ode_comp_string = None
        self._pde_comp_string = None
        self._custom_comp_string = None
        self.sigma = None
        self.lib_strings = None

    def init_library(self, ode_order, pde_order, indices=None, custom_terms=np.array([]), kind="split"):
        """
        Method that initializes all library strings up to specified order. These string will be pre-compiled and later
        evaluated in the right hand side method. They include products of all provided variabled up to order ode_order
        and their spatial derivatives up to order pde_order. Indices is used to filter only selected library terms for
        spatial differentiation.
        Parameters
        ----------
        ode_order : int
            Order up to which products of variables (including with themselves) are included in the library
        pde_order : int
            Order up to which spatial derivatives are to be computed
        indices : np.array
            (optional, default=None) Array of indices of library elements to be included for spatial derivation (as a
            subset of terms created undeer ode_order)
        custom_terms : np.array
            (optional, default=np.array([])) User defined string to append to library and evaluate in RHS
        kind : str
            (optional, default="split") Whether to split derivatives into each spatial direction and their combinations
            (exhaustive) or "pair" them each (non-exhaustive).

        Notes
        -----
        Library creation should focus on the array-valued attribute "lib_string" and later evaluation should use that
        array for easier manipulation of library terms. Then one wouldn't need all this bogus differentiation between
        ode, pde or custom strings.

        """
        self.ode_order = ode_order  # Maximum number of factors in products of permutations
        self.pde_order = pde_order  # Maximum order of spatial derivative
        self.ode_count = self._give_count(self.ode_order)
        if indices is None:
            self.indices_ode_to_pde = np.arange(1, self.ode_count)
        else:
            self.indices_ode_to_pde = indices
        self.pde_count = self.pde_order * self.dimension * len(self.indices_ode_to_pde)

        self.ode_string = _make_combination_string(self.ode_order, self.n_variables)
        self._ode_comp_string = compile(self.ode_string, '<string>', 'eval')

        ode_strings = self.ode_string.split(", ")
        ode_strings[-1] = ode_strings[-1][:-2]

        def make_split_pde_terms():
            x_string = "np.array(["
            for order in np.arange(1, pde_order + 1):
                for index in self.indices_ode_to_pde:
                    if self.dimension == 0:
                        x_string += "  "
                        break
                    # x_string += pre + f"self.diff.d_dx(ode_lib[{index}],{order})"
                    x_string += f"self.diff.d_dx({ode_strings[index]},{order}), "
                    if self.dimension == 2:
                        # x_string += f", self.diff.d_dy(ode_lib[{index}],{order})"
                        x_string += f"self.diff.d_dy({ode_strings[index]},{order}), "
                        combs = [c for c in combinations_wr(range(1, order + 1), 2) if sum(c) == order]
                        for comb in combs:
                            x_string += f"self.diff.dd_dxdy({ode_strings[index]},[{comb[0]},{comb[1]}]), "
                            self.pde_count += 1
                            if comb[0] != comb[1]:
                                x_string += f"self.diff.dd_dxdy({ode_strings[index]},[{comb[1]},{comb[0]}]), "
                                self.pde_count += 1
            x_string = x_string[:-2] + (pde_order == 0)*"([" + "])"
            return x_string

        def make_coupled_pde_terms():
            x_string = "np.array(["
            for order in np.arange(1, pde_order + 1):
                for index in self.indices_ode_to_pde:
                    if self.dimension == 0:
                        x_string += "  "
                        break
                    x_string += f"self.diff.div({ode_strings[index]},{order}), "
            x_string = x_string[:-2] + (pde_order == 0) * "([" + "])"
            return x_string

        if kind == "split":
            x = make_split_pde_terms()
        else:
            x = make_coupled_pde_terms()
        self.pde_string = x
        self._pde_comp_string = compile(self.pde_string, '<string>', 'eval')
        self.sigma = np.zeros((self.n_variables, (self.ode_count + self.pde_count)))

        self.lib_strings = np.concatenate((self.ode_string[10:-2].split(", "), self.pde_string[10:-2].split(", ")))
        self.add_library_terms(custom_terms)

    def add_library_terms(self, list_of_strings):
        """
        Method that adds custom python expressions as strings and adds them to the library of RHS terms.

        Parameters
        ----------
        list_of_strings : numpy.array
            Array of strings to be added, can be empty.

        """
        assert self.lib_strings is not None
        self.custom_count = len(list_of_strings)
        self.lib_strings = np.concatenate((self.lib_strings, list_of_strings))

        if len(list_of_strings) == 0:
            if self.dimension == 0:
                x_string = f"np.array([]).reshape(0, )"
            elif self.dimension == 1:
                x_string = f"np.array([]).reshape(0, {self.N})"
            elif self.dimension == 2:
                x_string = f"np.array([]).reshape(0, {self.N}, {self.N})"
        else:
            x_string = "np.array(["
            for string in list_of_strings:
                x_string += string + ", "

            x_string = x_string[:-2] + "])"

        self.custom_string = x_string
        self._custom_comp_string = compile(self.custom_string, '<string>', 'eval')
        self.sigma = np.zeros((self.n_variables, (self.ode_count + self.pde_count + self.custom_count)))

    def drop_library_terms(self, indices):
        """
        Drop terms of library at the indices specified in the argument. See 'print_library' for an output of library
        terms and their indices.

        Parameters
        ----------
        indices : list or array-like
            List of indices to be dropped from library

        Notes
        -----
        This is all way too cumbersome and should be redone (as described in the note of init_library)

        """

        assert self.lib_strings is not None
        n_drops = len(indices)
        indices = np.array(indices)
        n_ode_drops = np.sum(indices < self.ode_count)
        n_pde_drops = np.sum(np.logical_and(indices >= self.ode_count, indices < (self.ode_count + self.pde_count)))
        n_custom_drops = np.sum(np.logical_and(indices >= (self.ode_count + self.pde_count),
                                               indices < (self.ode_count + self.pde_count + self.custom_count)))
        self.ode_count -= n_ode_drops
        self.pde_count -= n_pde_drops
        self.custom_count -= n_custom_drops

        all_library_strings = self.ode_string.split(", ") + self.pde_string.split(", ") + self.custom_string.split(", ")
        new_library_strings = np.delete(np.array(all_library_strings), indices)

        new_ode_strings = new_library_strings[0:self.ode_count]
        new_pde_strings = new_library_strings[self.ode_count:(self.ode_count + self.pde_count)]
        new_custom_strings = new_library_strings[(self.ode_count + self.pde_count):]

        flags = [False, False, False]
        if len(new_ode_strings) == 0:
            new_ode_strings = [" "]
            flags[0] = True
        if len(new_pde_strings) == 0:
            new_pde_strings = [" "]
            flags[1] = True
        if len(new_custom_strings) == 0:
            new_custom_strings = [" "]
            flags[2] = True

        if new_ode_strings[0][0:10] != "np.array([":
            new_ode_strings[0] = "np.array([" + new_ode_strings[0]
        if new_pde_strings[0][0:10] != "np.array([":
            new_pde_strings[0] = "np.array([" + new_pde_strings[0]
        if new_custom_strings[0][0:10] != "np.array([":
            new_custom_strings[0] = "np.array([" + new_custom_strings[0]

        if new_ode_strings[-1][-2:] != "])":
            new_ode_strings[-1] += "])"
            if flags[0]:
                new_ode_strings[-1] += f".reshape(0, {self.N}, {self.N})"
        if new_pde_strings[-1][-2:] != "])":
            new_pde_strings[-1] += "])"
            if flags[1]:
                new_pde_strings[-1] += f".reshape(0, {len(self.x)}, {len(self.y)})"
        if new_custom_strings[-1][-2:] != "])":
            new_custom_strings[-1] += "])"
            if flags[2]:
                new_custom_strings[-1] += f".reshape(0, {self.N}, {self.N})"

        self.ode_string = ", ".join(new_ode_strings)
        self.pde_string = ", ".join(new_pde_strings)
        self.custom_string = ", ".join(new_custom_strings)
        self._ode_comp_string = compile(self.ode_string, '<string>', 'eval')
        self._pde_comp_string = compile(self.pde_string, '<string>', 'eval')
        self._custom_comp_string = compile(self.custom_string, '<string>', 'eval')
        self.sigma = np.zeros((self.n_variables, (self.ode_count + self.pde_count + self.custom_count)))
        self.lib_strings = np.delete(self.lib_strings, indices)

    def _give_count(self, order):
        """
        (Internal) helper function that gives the number of ode-type terms that will automatically be created.

        Parameters
        ----------
        order : int
            Highest order up to which products will be created

        Returns
        -------
        int
            Number of ode-type terms that will be created automatically.

        """
        count = int(sum([math.factorial(self.n_variables + o - 1) / math.factorial(o) /
                         math.factorial(self.n_variables - 1) for o in range(order + 1)]))
        return count

    def init_simulator(self, sig=None, sol=None, time=None):
        """
        Method that initializes a TimeSimulator class and object. This object is a child of 'AbstractTimesimulator' of
        the time_simulation submodule. It is extended with a method that provides a time series of the library terms
        given a time series.

        Parameters
        ----------
        sig : numpy.array
            (Optional, default=None) Vector of factors for linearly combining library terms in right hand side. If none
            the attribute self.sigma is used instead (which is initialized as all zeros).
        sol : numpy.array
            (optional, default=None) Time series data to give to the time series simulator, if None self.sol is used.
        time : numpy.array
            (optional, default=None) Time stepping data corresponding to solution, if None self.time is used.

        Yields
        -------
        TimeSimulator object in Model attributes
        """
        if sig is None:
            sigma = self.sigma
        else:
            sigma = sig
        if sol is None:
            sol = self.sol
        if time is None:
            time = self.time

        ode_string = self._ode_comp_string
        pde_string = self._pde_comp_string
        custom_string = self._custom_comp_string
        initial_state = self.initial_state
        domain = self.domain
        boundary_condition = self.boundary_condition
        differentiation_method = self.differentiation_method
        if self.dimension == 1:
            n_shape = (self.N,)
        elif self.dimension == 2:
            n_shape = (self.N, self.N)

        class TimeSimulator(AbstractTimeSimulator):
            def __init__(self):
                super().__init__(initial_state, domain=domain, bc=boundary_condition,
                                 diff=differentiation_method)
                self.sigma = sigma
                self.library_over_time = None
                self.sol_dot = None

            def rhs(self, t, u):
                ode_lib = eval(ode_string)
                pde_lib = eval(pde_string)
                custom_lib = eval(custom_string)
                vec = np.concatenate((ode_lib, pde_lib, custom_lib))
                u_next = [np.dot(s, [item.flatten() for item in vec]).reshape(n_shape) for s in self.sigma]

                return np.array(u_next)

            def create_library_time_series(self, ts=None):
                if ts is None:
                    assert self.sol is not None, "Must provide time series or perform time simulation."
                    time_series = self.sol
                else:
                    time_series = ts

                # Feels inefficient (and it is)
                library_over_time = []
                for var in range(time_series.shape[0]):
                    var_lib_over_time = []
                    for t in range(time_series.shape[1]):
                        u = time_series[:, t]
                        ode_lib = eval(ode_string)
                        pde_lib = eval(pde_string)
                        custom_lib = eval(custom_string)
                        vec = np.concatenate((ode_lib, pde_lib, custom_lib))
                        var_lib_over_time.append(vec)
                    var_lib_over_time = np.swapaxes(np.array(var_lib_over_time), 0, 1)
                    library_over_time.append(var_lib_over_time)

                self.library_over_time = np.array(library_over_time)
                return self.library_over_time

        self.simulator = TimeSimulator()

        # Usually doesn't do anything
        self.simulator.sol = sol
        self.simulator.time = time

    def print_library(self, sigma=None):
        """
        Print a readable table of library terms, their index and if available their linear factor.

        Parameters
        ----------
        sigma : numpy.array
            (optional, default=None) If not None, replace values of linear combination factors with values stored in
            this array.

        """
        if self.sigma is None:
            print(tabulate(zip(range(len(self.lib_strings)), self.lib_strings), headers=["Index", "Term"]))
        else:
            if sigma is not None:
                sigma_loc = sigma
            else:
                sigma_loc = self.sigma
            headers = ["Index", "Term"]
            for i in range(sigma_loc.shape[0]):
                headers.append(f"Value {i}")

            print(tabulate(zip(range(len(self.lib_strings)), self.lib_strings, *sigma_loc), headers=headers))

    def optimize_sigma(self, lamb=0.1, thres=0.01, error="BIC", **kwargs):
        """
        Optimize sigma using variation of training for sequencing threshold ridge regression. See Module `opimization`
        for details.

        Parameters
        ----------
        lamb : float
            (Optional, default=0.1) Sparsity knob to promote sparsity in regularisation

        thres : float
            (Optional, default=0.01) Threshold below which values are cut in sequence of regression

        error : string
            (Optional, default="BIC") Keyword argument to pass to optimization routine. Can either use BIC with
            integrated error or non-integrated error otherwise.

        """
        print("Generating test functions (this may take some time)...")
        sol_dot = self.simulator.get_sol_dot()
        lot = self.simulator.create_library_time_series()
        # sol_dot_flat = sol_dot.reshape((sol_dot.shape[0], -1))
        # lot_flat = lot.reshape((lot.shape[0], lot.shape[1], -1))

        sigma_opt, error = train_ridge_wip(lot, self.sol, sol_dot, self.time, lamb, thres, error=error, **kwargs)

        self.sigma = sigma_opt
        print("New Sigma set to: \n")

        self.print_library()



from .simulator import AbstractTimeSimulator
from .visualization import animate

from scipy.integrate._ivp.rk import OdeSolver  # this is the class we will monkey patch

from tqdm import tqdm

### monkey patching the ode solvers with a progress bar

# save the old methods - we still need them
old_init = OdeSolver.__init__
old_step = OdeSolver.step


# define our own methods
def new_init(self, fun, t0, y0, t_bound, vectorized, support_complex=False):
    # define the progress bar
    self.pbar = tqdm(total=t_bound - t0, unit='ut', initial=t0, ascii=True, desc='IVP')
    self.last_t = t0

    # call the old method - we still want to do the old things too!
    old_init(self, fun, t0, y0, t_bound, vectorized, support_complex)


def new_step(self):
    # call the old method
    old_step(self)

    # update the bar
    tst = self.t - self.last_t
    self.pbar.update(tst)
    self.last_t = self.t

    # close the bar if the end is reached
    if self.t >= self.t_bound:
        self.pbar.close()


# overwrite the old methods with our customized ones
OdeSolver.__init__ = new_init
OdeSolver.step = new_step



from setuptools import setup, find_packages


long_description = "The ``tsme`` package provides estimation methods for ODE or PDE systems using timeseries data. It also includes basic time simulation capabilities by wrapping scipy's initial value problem solver, either using finite differences or pseudo-spectral methods for spatial derivatives."

setup(
    name='tsme',
    version='0.0.3',
    url='https://zivgitlab.uni-muenster.de/nonlinear-physics/ag-kamps/tsme',
    license='GPL',
    author='Oliver Mai',
    author_email='oliver.mai@wwu.de',
    install_requires=['numpy', 'matplotlib', 'scipy', 'findiff', 'tabulate', 'tqdm', 'hyperopt', 'optuna', 'nevergrad'],
    scripts=[],
    packages=find_packages(),
    description='A package that provides estimation methods for differential equations of dynamical systems based on timeseries data.',
    long_description=long_description,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Physics'
        ])



# Brainstorming for an ModelEstimator Class, which takes a time series and provides methods for estimating a Model
# describing the time evolution
# this class needs to house a subclass "Model", which in turn needs to create a library of possible terms in the right
# hand side

